use strict;
package europa_module_factory;

use vars qw ($AUTOLOAD);

my $db = {};

# Object constructor.
sub new
{
  my $class = shift;
  my $options = shift;

  my $this = $options;
  bless($this, $class);

  my $first_module_of_its_class = 0;
  my $db_key = ref $this;
  if (!exists $db->{$db_key})
  {
    $db->{$db_key} = {};
    $db->{$db_key}->{fields} = $class->get_fields();
    $first_module_of_its_class = 1;
  }

  # Call setters with each option.
  for my $option (keys %$options)
  {
    my $value = $options->{$option};
    $this->$option($value);
  }

  # Verify that required fields were supplied.
  my $fields = $class->get_fields();
  for my $field (keys %{$fields})
  {
    if (!$fields->{$field}->{optional})
    {
      if (!defined $this->$field())
      {
        print STDERR "$class: required option '$field' is not defined\n";
        exit 1;
      }
    }
  }

  if (!defined $this->module())
  {
    # If no name is assigned, come up with a unique one based
    # on $class and options.  But if this is the first module
    # of its class (which may be a common case) take a shortcut -
    # choose a simple name.
    if (!defined $this->{name})
    {
      my $name = $class;
      $name =~ s/::/_/g;
      
      if (!$first_module_of_its_class)
      {
        $name .= "_" if (keys %$this);
        $name .= join("_", (map {"$_\_$this->{$_}"} keys %$this));
      }
      $this->{name} = $name;
    }
    my $module = e_module->new({name => $this->{name}});

    $this->module($module);
    $this->document();

    $this->add_contents_to_module();

    $this->test_ports();
  }

  return $this;
}

# Prevent autoload on this method, because I don't want it to
# have a setter.
# Derived class can override this method to declare port direction
# and width, and get automatic checking with warnings upon mismatch.
# 
# The method returns a reference to a hash of this form:
# 
# my %ports = (
#   <port name> => {
#     width => <port width>,
#     direction => <input|output>,
#   },
#   ...
# );
sub declare_ports
{
  return undef;
}

sub test_ports
{
  my $this = shift;
  my $ports = $this->declare_ports();

  return if (!defined $ports);

  my $warning_prefix = ref($this) . ": ";
  my %inputs = map {$_ => 1} $this->module()->get_input_names();
  my %outputs = map {$_ => 1} $this->module()->get_output_names();

  # For each declared port: if it's found as an actual port, delete
  # it from the hash.  Issue a warning for each remaining port.
  for my $port (keys %$ports)
  {
    if ($ports->{$port}->{direction} eq 'input')
    {
      if (defined $inputs{$port})
      {
        # Check the width.
        # To do: if the signal exists, but has the wrong width, right now
        # it's reported as a plain old undeclared port.  WOuld be better 
        # to report as a unexpected width port.
        if ($this->_get_signal_width($port) == $ports->{$port}->{width})
        {
          delete $ports->{$port};
          delete $inputs{$port};
        }
      }
    }
    elsif ($ports->{$port}->{direction} eq 'output')
    {
       if (defined $outputs{$port})
       {
        # Check the width.
        # To do: if the signal exists, but has the wrong width, right now
        # it's reported as a plain old undeclared port.  WOuld be better 
        # to report as a unexpected width port.
        if ($this->_get_signal_width($port) == $ports->{$port}->{width})
        {
          delete $ports->{$port};
          delete $outputs{$port};
        }
       }
    }
    else
    {
      print STDERR $warning_prefix . 
        "WARNING: Unknown direction '$port->{direction}' for port '$port'\n";
    }
  }

  if (keys %$ports)
  {
    print STDERR $warning_prefix . 
      "WARNING: declared ports not found in generated output:\n";
    for my $port (keys %$ports)
    {
      print STDERR
        "\tport:'$port'; " .
        "dir: $ports->{$port}->{direction}; " .
        "width: $ports->{$port}->{width}\n";
    }
  }

  if (keys %inputs)
  {
    print STDERR $warning_prefix .
      "WARNING: undeclared input ports found in generated output:\n";
    for my $name (sort keys %inputs)
    {
      my $desc = $this->_get_signal_description($name);
      print "\t$desc\n";
    }
  }
  if (keys %outputs)
  {
    print STDERR $warning_prefix .
      "WARNING: undeclared output ports found in generated output:\n";
    for my $name (sort keys %outputs)
    {
      my $desc = $this->_get_signal_description($name);
      print "\t$desc\n";
    }
  }
}

sub _get_signal_width
{
  my $this = shift;
  my $name = shift;
  my $sig = $this->module()->get_object_by_name($name);
  my $width = 1;
  my $is_default = 1;
  if ($sig)
  {
    $width = $sig->width();
    $is_default = 0;
  }

  if (wantarray)
  {
    return ($width, $is_default);
  }
  return $width;
}

sub _get_signal_description
{
  my $this = shift;
  my $name = shift;
  my ($width, $is_default) = $this->_get_signal_width($name);
  my $default_width_suffix = $is_default ? "*" : "";

  return "$name ($width$default_width_suffix)";
}

# AUTOLOAD setters and getters.
# If a fields hash is defined for the option name,
# it's used to verify the value.  No fields hash means
# anything goes... may want to make that be a failure instead.
sub AUTOLOAD
{
  my $this = shift;
  my $value = shift;
  my $name = $AUTOLOAD;

  # Strip off all but the last element of $name.  Separator is "::".
  my @parts = split('::', $name);
  $name = $parts[$#parts];

  # This method can be called statically or with an actual object.
  my $class = $this;
  $class = ref($class) if ref($class);

  if (defined $value)
  {
    # Look up this class's option validations.
    my $fields = $db->{$class}->{fields};
    if (!$class->validate($name, $value, $fields->{$name}))
    {
      exit 1;
    }
    if (ref $this)
    {
      $this->{$name} = $value;
    }
  }
  return $this->{$name};
}

# Subclass should override this method to add module contents.
# If the method is not overridden, an empty stub module is created.
#
# To do:Instances of the empty module appear to be empty - it would be 
# nice if they showed up, even if their existence meant bogus HDL.
sub add_contents_to_module
{
  my $this = shift;
  my $module = $this->module();

  my $existing_comment = $module->comment();
  $module->comment(
    $existing_comment . 
    "\nThis is a stub module.\nAdd contents to this module by overriding\n" . 
    ref($this) .
    "::add_contents_to_module().\n"
  );
}

sub validate
{
  my $class = shift;
  my $name = shift;
  my $value = shift;
  my $validator = shift;
  my $retval = 1;

  return 1 if (!defined $validator);
  if (defined $validator->{range})
  {
    my ($min, $max) = @{$validator->{range}};

    if (defined($min) && $value < $min)
    {
      print STDERR "$class\::validate(): " .
        "value '$value' for option '$name' is less than " .
        "allowed min value '$min'\n";
      $retval = 0;
    }
    if (defined($max) && $value > $max)
    {
      print STDERR "$class\::validate(): " .
        "value '$value' for option '$name' is greater than " .
        "allowed max value '$max'\n";
      $retval = 0;
    }
  }
  elsif (defined $validator->{allowed_values})
  {
    if (!grep {$value eq $_} @{$validator->{allowed_values}})
    {
       print STDERR "$class\::validate(): " .
        "value '$value' for option '$name' is not in allowed set\n" .
        join(", ", @{$validator->{allowed_values}}) .
        "\n";
      $retval = 0;
    }
  }

  return $retval;
}

sub get_existing_module
{
  my $this = shift;
  my $module = undef;

  my $key = $this->compute_module_key();
  if (exists $db->{ref $this}->{$key})
  {
    $module = $db->{ref $this}->{$key};
  }

  return $module;
}

sub compute_module_key
{
  my $this = shift;

  my $key = join("", (map {"$_=>$this->{$_},"} sort keys %$this));
  return $key;
}

sub module
{
  my $this = shift;
  if (@_) 
  {
    my $key = $this->compute_module_key();
    $db->{ref $this}->{$key} = shift;
  }

  return $this->get_existing_module();
}

sub document
{
  my $this = shift;

  my $comment;
  # Put the name of the instancing class right at the top, for reference.
  $comment .= ("Module class: " . ref($this) . "\n");
  $comment .= "Module options:\n";

  for my $key (sort keys %$this)
  {
    my $value = $this->{$key};
    $value = "<ref>" if ref $value;
    $comment .= "$key: $value\n";
  }

  my $module = $this->module();
  if (defined $module)
  {
    $this->module()->comment($comment);
  }
}

sub make_instance
{
  my $this = shift;

  my $options = shift;

  my $module = $this->get_existing_module();
  $options->{module} = $module;

  my $instance = undef;
  if (defined $module)
  {
    $instance = e_instance->new($options);
  }
  else
  {
    print STDERR __PACKAGE__ . "::make_instance(): Error: non-existent module\n";
  }

  return $instance;
}

1;

