use strict;
use europa_all;
use component_utils;

sub print_usage_and_fail
{
  my $msg = shift;
  print STDERR "\n$0: $msg\n";
  print STDERR 
    "\nUsage:\n" .
    "$0 [--help]\n" .
    "(Print this help message)\n\n" .
    "$0 --component=<component name> \\\n\t" .
    "--top=<top level module> \\\n\t" .
    "--help\n" .
    "(Print sub-package-specific (component::top) help)\n\n" .
    "$0 --component=<component name> \\\n\t" .
    "--top=<top level module> \\\n\t" .
    "--name=<top level module name> \\\n\t" .
    "<component-specific options>\n" .
    "(Create a module of type component::top, with the given name and options)\n\n";

  exit 1;
}

sub print_help_and_exit
{
  my $top_package = shift;
  my $fields = $top_package->get_fields();

  print STDERR "\nAllowed fields in package '$top_package':\n";
  for my $field (sort keys %$fields)
  {
    my $desc;
    if ($fields->{$field}->{description})
    {
      $desc = "\n\t" . $fields->{$field}->{description};
    }

    if (defined $fields->{$field}->{range})
    {
      my $min = $fields->{$field}->{range}->[0];
      $min = "minint" if !defined $min;
      my $max = $fields->{$field}->{range}->[1];
      $max = "maxint" if !defined $max;

      print "$field:$desc" . "\n\trange: [$min .. $max]\n";
    }
    elsif (defined $fields->{$field}->{allowed_values})
    {
      print "$field:$desc" . "\n\tallowed values: " .
        join(", ", @{$fields->{$field}->{allowed_values}}) .
        "\n";
    }
  }

  exit 1;
}

my $options = component_utils::get_args(@ARGV);

# Consume these parameters at this level; other parameters
# propagate downward.
#  component
#  top
#  help
#
# "name" is a funny special case: it's required at this level,
# but is not consumed at this level; instead, it's passed downward.

# Extract the component (perl package) to build.
my $component = $options->{component};
delete $options->{component};

my $top = $options->{top};
delete $options->{top};

my $name = $options->{name};

my $error;
$error = "\nMissing required parameter 'component'"
  if (!defined $component);
$error .= "\nMissing required parameter 'top'"
  if (!defined $top);
$error .= "\nMissing required parameter 'name'"
  if (!defined $name && !exists($options->{help}));
print_usage_and_fail($error) if defined $error;

my $top_perl_module = "$component/$top.pm";
eval {require $top_perl_module};
if ($@)
{
  print STDERR "Failed to load module '$top_perl_module'\n";
  exit 1;
}

# Extract the target directory from the options.
my $target_dir = $options->{target_dir};
delete $options->{target_dir};
$target_dir = '.' if !defined $target_dir;

my $top_package = $component;
if ($component ne $top)
{
  $top_package = "$component\::$top";
}

if (exists $options->{help})
{
  print_help_and_exit($top_package);
}

my $top_module = $top_package->new($options)->module();

my $project = e_project->new({
  do_write_ptf => 0,
  _system_directory => $target_dir,
});

$project->top($top_module);
$project->output();


